# -*- coding: utf-8 -*-
"""
Boolean Node
=============

Main class for Boolean node objects.

"""
#   Copyright (C) 2021 by
#   Rion Brattig Correia <rionbr@gmail.com>
#   Alex Gates <ajgates@gmail.com>
#   Etienne Nzabarushimana <enzabaru@indiana.edu>
#   All rights reserved.
#   MIT license.
from __future__ import division
import numpy as np
import pandas as pd
from statistics import mean
from itertools import compress, combinations
from cana.canalization import boolean_canalization as BCanalization
from cana.canalization import cboolean_canalization as cBCanalization
from cana.utils import *
from cana.cutils import *


class BooleanNode(object):
    """
    """

    def __init__(self, id=0, name='x', k=1, inputs=[1], state=False, outputs=[0, 1], constant=False, network=None, verbose=False, *args, **kwargs):
        self.id = id                            # the id of the node
        self.name = name                        # the name of the node
        self.k = k                              # k is the number of inputs
        self.inputs = list(map(int, inputs))    # the ids of the input nodes
        self.state = state                      # the initial state of the node
        self.outputs = list(map(str, outputs))  # the list of transition outputs
        self.network = network                  # the BooleanNetwork object this nodes belongs to
        self.verbose = verbose                  # verbose mode

        # mask for inputs
        if len(self.inputs) > 0:
            self.mask = [(i in self.inputs) for i in range(max(self.inputs) + 1)]
        else:
            self.mask = []

        # Consistency
        if (k != 0) and (k != int(np.log2(len(outputs)))):
            raise ValueError('Number of k (inputs) do not match the number of output transitions')

        # If all outputs are either positive or negative, this node can be treated as a constant.
        if (len(set(outputs)) == 1) or (constant):
            self.constant = True
            self.step = self.constant_step
        else:
            self.constant = False
            self.step = self.dynamic_step

        # Canalization Variables
        self._prime_implicants = None           # A tuple of negative and positive prime implicants.
        self._two_symbols = None                # The Two Symbol (TS) Schemata
        self._pi_coverage = None                # The Coverage of inputs by Prime Implicants schemata
        self._ts_coverage = None                # The Coverage of inputs by Two Symbol schemata

    def __str__(self):
        if len(self.outputs) > 10:
            outputs = '[' + ','.join(map(str, self.outputs[:4])) + '...' + ','.join(map(str, self.outputs[-4:])) + ']'
        else:
            outputs = '[' + ','.join(map(str, self.outputs)) + ']'
        inputs = '[' + ','.join(map(str, self.inputs)) + ']'
        return "<BNode(id={id:d}, name='{name:s}', k={k:d}, inputs={inputs:s}, state={state:d}, outputs={outputs:s} constant={constant:b})>".format(
            id=self.id, name=self.name, k=self.k, inputs=inputs, state=self.state, outputs=outputs, constant=self.constant)

    @classmethod
    def from_output_list(self, outputs=list(), *args, **kwargs):
        """Instanciate a Boolean Node from a output transition list.

        Args:
            outputs (list) : The transition outputs of the node.

        Returns:
            (BooleanNode) : the instanciated object.

        Example:
            >>> BooleanNode.from_output_list(outputs=[0,0,0,1], name="AND")
        """
        id = kwargs.pop('id') if 'id' in kwargs else 0
        name = kwargs.pop('name') if 'name' in kwargs else 'x'
        k = int(np.log2(len(outputs)))
        inputs = kwargs.pop('inputs') if 'inputs' in kwargs else [(x + 1) for x in range(k)]
        state = kwargs.pop('state') if 'state' in kwargs else False

        return BooleanNode(id=id, name=name, k=k, inputs=inputs, state=state, outputs=outputs, *args, **kwargs)

    def input_redundancy(self, operator=mean, norm=True):
        r"""The Input Redundancy :math:`k_{r}` is the mean number of unnecessary inputs (or ``#``) in the Prime Implicants Look Up Table (LUT).
        Since there may be more than one redescription schema for each input entry, the input redundancy is bounded by an upper and lower limit.


        .. math::

            k_{r}(x) = \frac{ \sum_{f_{\alpha} \in F} \Phi_{\theta:f_{\alpha} \in \Theta_{\theta}} (n^{\#}_{\theta} ) }{ |F| }

        where :math:`\Phi` is a function (:math:`min` or :math:`max`) and :math:`F` is the node LUT.

        Args:
            operator (function) : The operator to use while computing input redundancy for the node.
                Defaults to `statistics.mean`, but `min` or `max` can be also considered.
            norm (bool) : Normalized between [0,1].
                Use this value when comparing nodes with different input sizes. (Defaults to "True".)

                :math:`k^{*}_r(x) = \frac{ k_r(x) }{ k(x) }`.


        Returns:
            (float) : The :math:`k_r` value.

        Note:
            The complete mathematical description can be found in :cite:`Marques-Pita:2013`.

        See also:
            :func:`effective_connectivity`, :func:`input_symmetry`, :func:`edge_redundancy`.
        """
        # Canalization can only occur when k>= 2
        if self.k < 2:
            return 0.0

        self._check_compute_canalization_variables(pi_coverage=True)

        if not hasattr(operator, '__call__'):
            raise AttributeError('The operator you selected must be a function. Try "min", "statitics.mean", or "max".')

        redundancy = [operator([pi.count('#') for pi in self._pi_coverage[binstate]]) for binstate in self._pi_coverage]

        k_r = sum(redundancy) / 2**self.k

        if (norm):
            # Normalizes
            k_r = k_r / self.k

        return k_r

    def edge_redundancy(self, bound='mean'):
        r""" The Edge Redundancy :math:`r_{i}` is the mean number of unnecessary inputs (or ``#``) in the Prime Implicants Look Up Table (LUT) for that input.
        Since there may be more than one redescription schema for each input entry, the input redundancy is bounded by an upper and lower limit.

        .. math::

            r_i(x_i) = \frac{ \sum_{f_{\alpha} \in F} \Phi_{\theta:f_{\alpha} \in \Theta_{\theta}} (X^{\#}_{\theta_i} ) }{ |F| }

        where :math:`\Phi` is a function (:math:`min` or :math:`max`) and :math:`F` is the node LUT.

        Args:
            bound (string) : The bound to which compute input redundancy.
                Mode "input" accepts: ["lower", "mean", "ave", upper", "tuple"].
                Defaults to "mean".

        Returns:
            (list) : The list of :math:`r_i` for inputs.

        Note:
            The complete mathematical description can be found in :cite:Gates:2020`.

        See also:
            :func:`effective_connectivity`, :func:`input_symmetry`.
        """

        self._check_compute_canalization_variables(pi_coverage=True)

        redundancies = []
        # Generate a per input coverage
        # ex: {0: {'11': [], '10': [], '00': [], '01': []}, 1: {'11': [], '10': [], '00': [], '01': []}}
        # pi_edge_coverage = { input : { binstate: [ pi[input] for pi in pis ] for binstate,pis in self._pi_coverage.items() } for input in range(self.k) }
        pi_edge_coverage = cBCanalization.input_wildcard_coverage(self._pi_coverage)
        # Loop ever input node
        for edge, binstates2wildcard in pi_edge_coverage.items():
            # {'numstate': [matches], '10': [True,False,True,...] ...}

            # countslenghts = {binstate_to_statenum(binstate): ([pi=='#' for pi in pis]) for binstate,pis in binstates.items() }
            # A triplet of (min, mean, max) values
            if bound == 'lower':
                redundancy = sum([all(pi) for pi in binstates2wildcard.values()]) / 2**self.k  # min(r_i)
            elif bound == 'mean' or bound == 'avg':
                redundancy = sum([sum(pi) / len(pi) for pi in binstates2wildcard.values()]) / 2**self.k  # <r_i>
            elif bound == 'upper':
                redundancy = sum([any(pi) for pi in binstates2wildcard.values()]) / 2**self.k  # max(r_i)
            elif bound == 'tuple':
                redundancy = (sum([all(pi) for pi in binstates2wildcard.values()]) / 2**self.k, sum([any(pi) for pi in binstates2wildcard.values()]) / 2**self.k)  # (min,max)
            else:
                raise AttributeError('The bound you selected does not exist. Try "upper", "mean", "lower" or "tuple".')

            redundancies.append(redundancy)

        return redundancies  # r_i

    def effective_connectivity(self, operator=mean, norm=True):
        r"""The Effective Connectiviy is the mean number of input nodes needed to determine the transition of the node.

        .. math::

            k_e(x) = k(x) - k_r(x)

        Args:
            operator (function) : The operator to use while computing input redundancy for the node.
                Defaults to `statistics.mean`, but `min` or `max` can be also considered.
            norm (bool) : Normalized between [0,1].
                Use this value when comparing nodes with different input sizes. (Defaults to "True".)

                :math:`k^{*}_e(x) = \frac{ k_e(x) }{ k(x) }`.

        Returns:
            (float) : The :math:`k_e` value.

        See Also:
            :func:`input_redundancy`, :func:`input_symmetry`, :func:`~cana.boolean_network.BooleanNetwork.effective_graph`.
        """
        #
        # Canalization can only occur when k>= 2
        if self.k < 2:
            return 0.0

        k_r = self.input_redundancy(operator=operator, norm=False)
        #
        k_e = self.k - k_r
        if (norm):
            k_e = k_e / self.k
        return k_e

    def edge_effectiveness(self, bound='mean'):
        r"""The Edge Effectiveness is the mean number of an input's states needed to determine the transition of the node.

        .. math::

            e_i(x_i) = 1 - r_i(x_i)

        Args:
            bound (string) : The bound for the :math:`k_r` Input Redundancy

        Returns:
            (list) : The list of :math:`e_r` values.

        See Also:
            :func:`input_redundancy`, :func:`input_symmetry`, :func:`~cana.boolean_network.BooleanNetwork.effective_graph`.
        """
        e_i = [1.0 - x_i for x_i in self.edge_redundancy(bound=bound)]
        return e_i

    def edge_symmetry(self, bound='upper'):
        r"""Edge Symmetry is a measure of permutation redundancy of a single input.
        Similar to the computation of Edge Effectiveness but using the Two-Symbol instead of the Prime Implicant schemata.

        .. math::

            s_i = \frac{ \sum_{f_{\alpha} \in F} \Phi_{\theta:f_{\alpha} \in \Theta_{\theta}} (n^{\circ}_i)}{ |F| }

        where :math:`\Phi` is the function :math:`min` or :math:`max` and :math:`F` is the node LUT.

        Args:
            bound (string) : The bound to which compute input symmetry.
                Mode "node" accepts: ["lower", "upper"].
                Mode "input" accepts: ["lower", "mean", "upper", "tuple"].
                Defaults to "upper".

        Returns:
            (float/list) : The :math:`k_s` or a list of :math:`r_i`.

        See also:
            :func:`input_redundancy`, :func:`effective_connectivity`
        """
        # Canalization can only occur when k>= 2
        if self.k < 2:
            return [0.0]

        self._check_compute_canalization_variables(ts_coverage=True)

        symmetries = []
        # Generate a per input coverage
        # ex: {0: {'11': [], '10': [], '00': [], '01': []}, 1: {'11': [], '10': [], '00': [], '01': []}}
        # ts_input_coverage = { input : { binstate: [ idxs.count(input) for schema,reps,sms in tss for idxs in reps+sms ] for binstate,tss in self._ts_coverage.items() } for input in range(self.k) }
        ts_input_coverage = {input: {binstate: [len(idxs) if input in idxs else 0 for schema, reps, sms in tss for idxs in reps + sms] for binstate, tss in self._ts_coverage.items()} for input in range(self.k)}

        # Loop ever input node
        for input, binstates in ts_input_coverage.items():
            # {'numstate': [number-of-ts's for each match], '10': [0, 2] ...}
            numstates = {binstate_to_statenum(binstate): permuts for binstate, permuts in binstates.items()}

            # A triplet of (min, mean, max) values
            if bound in ['lower', 'mean', 'upper']:
                # Min, Mean or Max
                if bound == 'upper':
                    minmax = max
                elif bound == 'mean':
                    minmax = np.mean
                elif bound == 'lower':
                    minmax = min

                s_i = sum(minmax(permuts) if len(permuts) else 0 for permuts in numstates.values()) / 2**self.k  # min(r_s)

            elif bound == 'tuple':
                # tuple (min,max) per input, per state
                s_i = [(min(permuts), max(permuts)) if len(permuts) else (0, 0) for permuts in numstates.values()]  # (min,max)
            else:
                raise AttributeError('The bound you selected does not exist. Try "upper", "mean", "lower" or "tuple".')
            symmetries.append(s_i)

        return symmetries  # s_i

    def input_symmetry(self, bound='upper', norm=True):
        r"""The Input Symmetry is a measure of permutation redundancy.
        Similar to the computation of Input Redundancy but using the Two-Symbol instead of the Prime Implicant schemata.

        .. math::

            k_s = \frac{ \sum_{f_{\alpha} \in F} \Phi_{\theta:f_{\alpha} \in \Theta_{\theta}} (n^{\circ}) }{ |F| }

        where :math:`\Phi` is the function :math:`min` or :math:`max` and :math:`F` is the node LUT.

        Args:
            bound (string) : The bound to which compute input symmetry.
                Mode "node" accepts: ["lower", "upper"].
                Mode "input" accepts: ["lower", "mean", "upper", "tuple"].
                Defaults to "upper".
            norm (bool) : Normalized between [0,1].
                Use this value when comparing nodes with different input sizes. (Defaults to "True".)

                :math:`k^{*}_s(x) = \frac{ k_s(x) }{ k(x) }`.

        Returns:
            (float/list) : The :math:`k_s` or a list of :math:`r_i`.

        See also:
            :func:`input_redundancy`, :func:`effective_connectivity`
        """
        # Canalization can only occur when k>= 2
        if self.k < 2:
            return 0.0

        self._check_compute_canalization_variables(ts_coverage=True)

        k_s = sum(self.edge_symmetry(bound=bound)) / self.k

        if (norm):
            k_s = k_s / self.k

        return k_s

    def look_up_table(self):
        """ Returns the Look Up Table (LUT)

        Returns:
            (pandas.DataFrame): the LUT

        Examples:
            >>> AND = BooleanNode.from_output_list([0,0,0,1])
            >>> AND.look_up_table()

        See also:
            :func:`schemata_look_up_table`

        """
        d = []
        if ((self.k < 2) and (self.constant is False)):
            k = 2
        else:
            k = self.k

        for statenum, output in zip(range(2**k), self.outputs):
            # Binary State, Transition
            d.append((statenum_to_binstate(statenum, base=self.k), output))

        df = pd.DataFrame(d, columns=['In:', 'Out:'])

        return df

    def schemata_look_up_table(self, type='pi', pi_symbol=u'#', ts_symbol_unicode=u"\u030A", ts_symbol_latex=u"\circ", format='pandas'):
        """ Returns the simplified schemata Look Up Table (LUT)

        Args:
            type (string) : The type of schemata to return, either Prime Implicants ``pi`` or Two-Symbol ``ts``. Defaults to 'pi'.
            pi_symbol (unicode) : The Prime Implicant don't care symbol. Default is ``#``.
            ts_symbol_unicode (unicode) : A unicode string for the Two Symbol permutable symbol. Default is ``u"\u030A"``.
            ts_symbol_latex (unicode) : The latex string for Two Symbol permutable symbol. Default is ``\circ``.
            format (string) : The format to return. Possible values are ``pandas`` (default) and ``latex``.

        Returns:
            (pandas.DataFrame or Latex): the schemata LUT

        Examples:
            >>> AND = BooleanNode.from_output_list([0,0,0,1])
            >>> AND.schemata_look_up_table(type='pi')

        Note:
            See the full list of `combining characters <https://en.wikipedia.org/wiki/Combining_character>`_ to use other symbols as the permutation symbol.

        See also:
            :func:`look_up_table`
        """
        r = []
        # Prime Implicant LUT
        if type == 'pi':
            self._check_compute_canalization_variables(prime_implicants=True)

            pi0s = self._prime_implicants.get('0', [])
            pi1s = self._prime_implicants.get('1', [])

            for output, pi in zip([0, 1], [pi0s, pi1s]):
                for schemata in pi:
                    r.append((schemata, output))

        # Two Symbol LUT
        elif type == 'ts':
            self._check_compute_canalization_variables(two_symbols=True)

            ts0s, ts1s = self._two_symbols

            for output, ts in zip([0, 1], [ts0s, ts1s]):
                for i, (schemata, permutables, samesymbols) in enumerate(ts):
                    string = u''
                    if len(permutables):
                        # Permutable
                        for j, permutable in enumerate(permutables):

                            if format == 'latex':
                                if j > 0:
                                    string += u' \,  | \, '
                                string += r' '.join([x if (k not in permutable) else '\overset{%s}{%s}' % (ts_symbol_latex, unicode(x)) for k,x in enumerate(schemata, start=0)])
                            else:
                                if j > 0:
                                    string += u' | '
                                string += u''.join([x if (k not in permutable) else u'%s%s' % (x, ts_symbol_unicode) for k, x in enumerate(schemata, start=0)])

                    else:
                        string += schemata
                    """
                    if len(samesymbols):
                        # Same Symbol
                        for j,samesymbol in enumerate(samesymbols):
                            if j>0:
                                sstring+= ' | '
                            sstring += ''.join([x if (k not in samesymbol) else unicode(x)+ts_symbols[-1] for k,x in enumerate(schemata, start=0)])
                    """
                    r.append((string, output))
        else:
            raise AttributeError('The schemata type could not be found. Try "PI" (Prime Implicants) or "TS" (Two-Symbol).')

        # Output Format (Latex Table or Pandas DataFrame)
        if format == 'latex':
            out = r"\begin{array}{ | c | r | l }" + "\n"
            out += r"\hline" + "\n"
            out += r" & In: & Out: \\" + "\n"
            out += r"\hline" + "\n"
            for i, (string, output) in enumerate(r):
                string = string.replace('2','\%s' % (pi_symbol))
                out += r"%d & %s & %s \\" % (i, string, output) + r"\hline" + "\n"
            out += r"\hline" + "\n"
            out += r"\end{array}" + "\n"
            return out

        elif format == 'pandas':

            r = [(schemata.replace('2', pi_symbol), output) for schemata, output in r]
            return pd.DataFrame(r, columns=['In:', 'Out:'])

        else:
            AttributeError('The format type could not be found. Try "pandas" "latex".')

    def input_mask(self, binstate):
        """ Returns the mask applied to the binary state binstate

        Args:
            binstate (str) : the binary state

        Returns:
            output (str) : the masked state
        """
        return ''.join(compress(binstate, self.mask))

    def constant_step(self, input_state):
        """
            Treat the node as a constant
        """
        return self.outputs[0]

    def dynamic_step(self, input_state):
        """ Returns the output of the node based on a specific input

        Args:
            input (list) : an input to the node.

        Returns:
            output (bool) : the output value.
        """
        return self.outputs[binstate_to_statenum(input_state)]

    def activities(self):
        return self.effective_connectivity(mode='input', bound='upper')

    def canalizing_map(self, output=None):
        """ Computes the node Canalizing Map (CM).

        Args:
            output (int) : The output CM to return. Default is ``None``, retuning both [0,1].

        Returns:
            CM (networkx.DiGraph) : a directed graph representation of the CM.

        See Also:
            :func:`boolean_network.dynamics_canalization_map` for the DCM and :func:`drawing.draw_canalizing_map_graphviz` for plotting.
        """
        self._check_compute_canalization_variables(two_symbols=True)

        ts0s, ts1s = self._two_symbols

        G = nx.DiGraph(name='CM: %s' % self.name)

        # Outputs
        if output is None or output == 0:
            G.add_node('var-{nid:d}-out-0'.format(nid=self.id), **{'label': self.name, 'type': 'variable', 'mode': 'output', 'value': 0, 'constant': self.constant, 'group': self.id})

        if output is None or output == 1:
            G.add_node('var-{nid:d}-out-1'.format(nid=self.id), **{'label': self.name, 'type': 'variable', 'mode': 'output', 'value': 1, 'constant': self.constant, 'group': self.id})

        tid = 0
        for out, tspsss in zip([0, 1], self._two_symbols):

            # Only return the requested output
            if ((not len(tspsss)) or ((output is not None) and (output != out))):
                continue

            for ts, ps, ss in tspsss:

                lits = []
                group0 = []
                group1 = []
                # group2 = []
                # Tau is the threshold, counted as the sum of (0's and 1's literals; 0's in permutation group; 1's in permutation group)
                nlit, ngrp0, ngrp1 = 0, 0, 0

                for j in range(self.k):
                    # Is this input in any permutation group?
                    input = ts[j]
                    if not any([j in group for group in ps]):
                        if ts[j] in ['0', '1']:
                            nlit += 1
                            source = j
                            lits.append(source)
                    else:
                        if ts[j] == '0':
                            ngrp0 += 1
                            group0.append(j)
                        elif ts[j] == '1':
                            ngrp1 += 1
                            group1.append(j)

                tau = nlit + ngrp0 + ngrp1

                # Threshold Node
                tname = 'thr-{tid:d}-var-{nid:d}-out-{out:d}'.format(tid=tid, nid=self.id, out=out)
                label = "{:d}".format(tau)
                G.add_node(tname, **{'label': label, 'type': 'threshold', 'tau': tau, 'group': self.id})

                # Add Edges from Threshold node to output
                xname = 'var-{nid:d}-out-{out:d}'.format(nid=self.id, out=out)
                if G.has_node(tname) and G.has_node(xname):
                    G.add_edge(tname, xname, **{'type': 'out'})
                else:
                    raise BaseException("Adding edge to node that does not exist.")

                # Literal Edges
                for lit in lits:
                    iname = 'var-{nid:d}-out-{out:d}'.format(nid=self.inputs[lit], out=int(ts[lit]))
                    if iname not in G.nodes():
                        # Can we get the name of ther nodes? Are we attached to a network?
                        if self.network is not None:
                            ilabel = self.network.get_node_name(self.inputs[lit])[0]
                        else:
                            ilabel = str(self.inputs[lit])
                        iout = int(ts[lit])
                        G.add_node(iname, **{'label': ilabel, 'type': 'variable', 'mode': 'input', 'value': iout, 'group': self.id})
                    G.add_edge(iname, tname, **{'type': 'literal'})

                # Group0
                for fusion in range(ngrp0):
                    fname = 'fus-{fus:d}-thr-{thr:d}-var-{nid:d}-out-{out:d}'.format(fus=fusion, thr=tid, nid=self.id, out=0)
                    G.add_node(fname, **{'type': 'fusion', 'group': self.id})
                    for input in ps[0]:
                        iname = 'var-{nid:d}-out-{out:d}'.format(nid=self.inputs[input], out=0)
                        if iname not in G.nodes():
                            # Can we get the name of ther nodes? Are we attached to a network?
                            if self.network is not None:
                                ilabel = self.network.get_node_name(self.inputs[input])[0]
                            else:
                                ilabel = str(self.inputs[input])
                            G.add_node(iname, **{'label': ilabel, 'type': 'variable', 'mode': 'input', 'value': 0, 'group': self.id})
                        G.add_edge(iname, fname, **{'type': 'fusing'})
                    G.add_edge(fname, tname, **{'type': 'fused'})

                # Group1
                for fusion in range(ngrp1):
                    fname = 'fus-{fus:d}-thr-{thr:d}-var-{nid:d}-out-{out:d}'.format(fus=fusion, thr=tid, nid=self.id, out=1)
                    G.add_node(fname, **{'type': 'fusion', 'group': self.id})
                    for input in ps[0]:
                        iname = 'var-{var:d}-out-{out:d}'.format(var=self.inputs[input], out=1)
                        if iname not in G.nodes():
                            # Can we get the name of ther nodes? Are we attached to a network?
                            if self.network is not None:
                                ilabel = self.network.get_node_name(self.inputs[input])[0]
                            else:
                                ilabel = str(self.inputs[input])
                            iout = ts[input]
                            G.add_node(iname, **{'label': ilabel, 'type': 'variable', 'mode': 'input', 'value': 1, 'group': self.id})
                        G.add_edge(iname, fname, **{'type': 'fusing'})
                    G.add_edge(fname, tname, **{'type': 'fused'})

                tid += 1

        return G

    def pi_coverage(self):
        """ Returns the :math:`F'` (Prime Implicants) binary state coverage.

        Returns:
            (list)

        See also:
            :func:`ts_coverage`
        """
        self._check_compute_canalization_variables(pi_coverage=True)
        return self._pi_coverage

    def ts_coverage(self):
        """ Returns the :math:`F''` (Two-Symbol schematas) binary state coverage.

        Returns:
            (list)

        See also:
            :func:`pi_coverage`
        """
        self._check_compute_canalization_variables(ts_coverage=True)
        return self._ts_coverage

    def _check_compute_canalization_variables(self, **kwargs):
        """ Recursevely check if the requested canalization variables are instantiated/computed, otherwise computes them in order.
        For example: to compute `two_symbols` we need `prime_implicants` first.
        Likewise, to compute `prime_implicants` we need the `transition_density_table` first.
        """
        if 'prime_implicants' in kwargs:
            if self._prime_implicants is None:
                self._prime_implicants = dict()
                for output in set(self.outputs):
                    output_binstates = outputs_to_binstates_of_given_type(self.outputs, output=output, k=self.k)
                    self._prime_implicants[output] = cBCanalization.find_implicants_qm(input_binstates=output_binstates)

        elif 'pi_coverage' in kwargs:
            self._check_compute_canalization_variables(prime_implicants=True)
            if self._pi_coverage is None:
                self._pi_coverage = dict()
                for output, piset in self._prime_implicants.items():
                    self._pi_coverage.update(cBCanalization.return_pi_coverage(piset))

                # make sure every inputstate was covered by at least one prime implicant
                assert len(self._pi_coverage) == 2**self.k

        elif 'two_symbols' in kwargs:
            self._check_compute_canalization_variables(prime_implicants=True)
            if self._two_symbols is None:
                # this is a temporary fix until we update 'find_two_symbols' to cython.
                if '0' in self._prime_implicants:
                    pi0 = self._prime_implicants['0']
                    pi0 = set([pi.replace('#', '2') for pi in pi0])
                else:
                    pi0 = []
                if '1' in self._prime_implicants:
                    pi1 = self._prime_implicants['1']
                    pi1 = set([pi.replace('#', '2') for pi in pi1])
                else:
                    pi1 = []
                # /end fix
                self._two_symbols = \
                    (
                        BCanalization.find_two_symbols_v2(k=self.k, prime_implicants=pi0),
                        BCanalization.find_two_symbols_v2(k=self.k, prime_implicants=pi1)
                    )
        elif 'ts_coverage' in kwargs:
            self._check_compute_canalization_variables(two_symbols=True)
            if self._ts_coverage is None:
                self._ts_coverage = BCanalization.computes_ts_coverage(self.k, self.outputs, self._two_symbols)

        else:
            raise Exception('Canalization variable name not found. %s' % kwargs)
        return True

    def bias(self):
        r""" The node bias. The sum of the boolean output transitions divided by the number of entries (:math:`2^k`) in the LUT.

        .. math::

            bias(x) = \frac{ \sum_{f_{\alpha}\in F} s_{\alpha} }{ |F| }

        Returns:
            (float)

        See Also:
            :func:`~cana.boolean_network.BooleanNetwork.network_bias`
        """
        return sum(map(int, self.outputs)) / 2**self.k

    def c_sensitivity(self, c, mode="default", max_k=0):
        """ Node c-sensitivity.
        c-sensitivity is defined as: the mean probability that changing exactly
        ``c`` variables in input variables would change output value.
        There is another mode "forceK", which will be used to calculate Derrida value.
        In that mode, it would assume the number of input variables is specified as max_k
        this methods is equivalent to Derrida value in :cite:`Kadelka:2017`, only move a normalization
        coefficient from expression of Derrida value to c-sensitivity to simplify it

        Args:
            c (int) : the ``c`` in the definition of c-senstivity above
            mode (string) : either "default" or "forceK"
            max_k (int) : you must specify max_k when you set mode as 'forceK'

        Returns:
            (float)

        See Also:
            :func:`~cana.boolean_network.derrida_curve`
        """
        S_c_f = 0
        ic = min(c, self.k)

        if mode != 'forceK':
            for j in product('01', repeat=self.k):
                origin_config = list(j)
                for mut in combinations(range(self.k), ic):
                    mut_config = origin_config[:]
                    for i_mut in mut:
                        mut_config[i_mut] = flip_bit(mut_config[i_mut])
                    if self.step(''.join(origin_config)) != self.step(''.join(mut_config)):
                        S_c_f += 1
            if S_c_f == 0:
                return 0.
            return S_c_f / float(ncr(self.k, c)) / float(2 ** self.k)
        else:
            assert max_k >= self.k
            for ic in range(max(1, c + self.k - max_k), min(c, self.k) + 1):
                for j in product('01', repeat=self.k):
                    origin_config = list(j)
                    for mut in combinations(range(self.k), ic):
                        mut_config = origin_config[:]
                        for i_mut in mut:
                            mut_config[i_mut] = flip_bit(mut_config[i_mut])
                        if self.step(''.join(origin_config)) != self.step(
                                ''.join(mut_config)):
                            S_c_f += ncr(max_k - self.k, c - ic)
            return S_c_f / float(ncr(max_k, c)) / float(2 ** self.k)



__package__ = 'cana'
__title__ = u'CANAlization: Control & Redundancy in Boolean Networks'
__description__ = u'This package implements a series of methods used to study control, canalization and redundancy in Boolean networks.'

__author__ = """\n""".join([
    'Rion Brattig Correia <rionbr@gmail.com>',
    'Alex Gates <ajgates42@gmail.com>',
    'Xuan Wang <xw47@indiana.edu>',
    'Thomas Parmer <tjparmer@indiana.edu>',
    'Etienne Nzabarushimana <enzabaru@indiana.edu>',
    'Luis M. Rocha <rocha@indiana.edu>'
])

__copyright__ = u'2021, Correia, R. B., Gates, A., Rocha, L. M.'

__version__ = '0.1.1'
__release__ = '0.1.1'
#
__all__ = ['boolean_network', 'boolean_node']



# -*- coding: utf-8 -*-
"""
Biological Boolean Networks
=================================

A series of biological Boolean networks that can be directly loaded for experimentation.


"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates42@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   Xuan Wang <xw47@indiana.edu>
#   Thomas Parmer <tjparmer@indiana.edu>
#   All rights reserved.
#   MIT license.
import os
from .. boolean_network import BooleanNetwork


_path = os.path.dirname(os.path.realpath(__file__))
""" Make sure we know what the current directory is """


def THALIANA():
    """Boolean network model of the control of flower morphogenesis in Arabidopsis thaliana

    The network is defined in :cite:`Chaos:2006`.

    Returns:
        (BooleanNetwork)
    """
    return BooleanNetwork.from_file(_path + '/thaliana.txt', name="Arabidopsis Thaliana", keep_constants=True)


def DROSOPHILA(cells=1):
    """Drosophila Melanogaster boolean model.
    This is a simplification of the original network defined in :cite:`Albert:2008`.
    In the original model, some nodes receive inputs from neighboring cells.
    In this single cell network, they are condensed (nhhnHH) and treated as constants.

    There is currently only one model available, where the original neighboring cell signals are treated as constants.

    Args:
        cells (int) : Which model to return.

    Returns:
        (BooleanNetwork)
    """
    if cells == 1:
        return BooleanNetwork.from_file(_path + '/drosophila_single_cell.txt', name="Drosophila Melanogaster", keep_constants=True)
    else:
        raise AttributeError('Only single cell drosophila boolean model currently available.')


def BUDDING_YEAST():
    """

    The network is defined in :cite:`Fangting:2004`.

    Returns:
        (BooleanNetwork)
    """
    return BooleanNetwork.from_file(_path + '/yeast_cell_cycle.txt', name="Budding Yeast Cell Cycle", keep_constants=True)


def MARQUESPITA():
    """Boolean network used for the Two-Symbol schemata example.

    The network is defined in :cite:`Marques-Pita:2013`.

    Returns:
        (BooleanNetwork)
    """
    return BooleanNetwork.from_file(_path + '/marques-pita_rocha.txt', name="Marques-Pita & Rocha", keep_constants=True)


def LEUKEMIA():
    """Boolean network model of survival signaling in T-LGL leukemia

    The network is defined in :cite:`Zhang:2008`.

    Returns:
        (BooleanNetwork)
    """
    return BooleanNetwork.from_file(_path + '/leukemia.txt', type='logical', name="T-LGL Leukemia", keep_constants=True)


def BREAST_CANCER():
    """Boolean network model of signal transduction in ER+ breast cancer

    The network is defined in :cite:`Zanudo:2017`.

    Returns:
        (BooleanNetwork)
    """
    return BooleanNetwork.from_file(_path + '/breast_cancer.txt', type='logical', name="ER+ Breast Cancer", keep_constants=True)


_cell_collective_models = [
    'Apoptosis Network',
    'Arabidopsis thaliana Cell Cycle',
    'Aurora Kinase A in Neuroblastoma',
    'B bronchiseptica and T retortaeformis coinfection',
    'B cell differentiation',
    'Bordetella bronchiseptica',
    'Bortezomib Responses in U266 Human Myeloma Cells',
    'BT474 Breast Cell Line Long-term ErbB Network',
    'BT474 Breast Cell Line Short-term ErbB Network',
    'Budding Yeast Cell Cycle 2009',
    'Budding Yeast Cell Cycle',
    'Cardiac development',
    'CD4 T cell signaling',
    'CD4+ T Cell Differentiation and Plasticity',
    'CD4+ T cell Differentiation',
    'Cell Cycle Transcription by Coupled CDK and Network Oscillators',
    'Cholesterol Regulatory Pathway',
    'Colitis-associated colon cancer',
    'Cortical Area Development',
    'Death Receptor Signaling',
    'Differentiation of T lymphocytes',
    'EGFR & ErbB Signaling',
    'FA BRCA pathway',
    'Fanconi anemia and checkpoint recovery',
    'FGF pathway of Drosophila Signalling Pathways',
    'Glucose Repression Signaling 2009',
    'Guard Cell Abscisic Acid Signaling',
    'HCC1954 Breast Cell Line Long-term ErbB Network',
    'HCC1954 Breast Cell Line Short-term ErbB Network',
    'HGF Signaling in Keratinocytes',
    'HH Pathway of Drosophila Signaling Pathways',
    'HIV-1 interactions with T Cell Signalling Pathway',
    'Human Gonadal Sex Determination',
    'IGVH mutations in chronic lymphocytic leukemia',
    'IL-1 Signaling',
    'IL-6 Signalling',
    'Influenza A Virus Replication Cycle',
    'Iron acquisition and oxidative stress response in aspergillus fumigatus',
    'Lac Operon',
    'Lymphoid and myeloid cell specification and transdifferentiation',
    'Lymphopoiesis Regulatory Network',
    'Mammalian Cell Cycle 2006',
    'Mammalian Cell Cycle',
    'MAPK Cancer Cell Fate Network',
    'Metabolic Interactions in the Gut Microbiome',
    'Neurotransmitter Signaling Pathway',
    'Oxidative Stress Pathway',
    'PC12 Cell Differentiation',
    'Predicting Variabilities in Cardiac Gene',
    'Pro-inflammatory Tumor Microenvironment in Acute Lymphoblastic Leukemia',
    'Processing of Spz Network from the Drosophila Signaling Pathway',
    'Regulation of the L-arabinose operon of Escherichia coli',
    'Senescence Associated Secretory Phenotype',
    'Septation Initiation Network',
    'Signal Transduction in Fibroblasts',
    'Signaling in Macrophage Activation',
    'Signaling Pathway for Butanol Production in Clostridium beijerinckii NRRL B-598',
    'SKBR3 Breast Cell Line Long-term ErbB Network',
    'SKBR3 Breast Cell Line Short-term ErbB Network',
    'Stomatal Opening Model',
    'T cell differentiation',
    'T Cell Receptor Signaling',
    'T-Cell Signaling 2006',
    'T-LGL Survival Network 2008',
    'T-LGL Survival Network 2011 Reduced Network',
    'T-LGL Survival Network 2011',
    'TOL Regulatory Network',
    'Toll Pathway of Drosophila Signaling Pathway',
    'Treatment of Castration-Resistant Prostate Cancer',
    'Trichostrongylus retortaeformis',
    'Tumour Cell Invasion and Migration',
    'VEGF Pathway of Drosophila Signaling Pathway',
    'Wg Pathway of Drosophila Signalling Pathways',
    'Yeast Apoptosis']


def load_cell_collective_model(name=None):
    """Loads one of the Cell Collective :cite:`Helikar:2012` models.
    Models collected on Aug 2020.

    Args:
        name (str): the name of the model to be loaded.
            Accepts: ["Apoptosis Network", "Arabidopsis thaliana Cell Cycle", "Aurora Kinase A in Neuroblastoma", ...,
            "Wg Pathway of Drosophila Signalling Pathways", "Yeast Apoptosis"]

    Returns:
        (BooleanNetwork)

    Note:
        See source code for full list of models. Credits to Xuan Wang for compiling these models.
        We are working on making a Cell Collective direct loader. 
    """

    #
    if name not in _cell_collective_models:
        models_str = "'" + "','".join(_cell_collective_models) + "'"
        raise TypeError('Model name could not be found. Please specify one of the following models: {models:s}.'.format(models=models_str))
    else:
        return BooleanNetwork.from_file(_path + '/cell_collective/' + name + '.txt', name=name, keep_constants=True)


def load_all_cell_collective_models():
    """Load all the Cell Collective models, instanciating +70 models.

    Returns:
        (list)

    Note:
        See source code for full list of models.
    """
    return [load_cell_collective_model(name=name) for name in _cell_collective_models]



# -*- coding: utf-8 -*-
"""
Network Motifs
===============

Simple network motifs in Networkx.DiGraph format that can be directly loaded.

"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   All rights reserved.
#   MIT license.
import networkx as nx


def network_motif(name=None):
    """Graph motifs from :cite:`Milo:2012`.

    Args:
        name (string): The name of the motif.
            Possible values are : ``FeedForward``, ``Fan``, ``FeedForwardSelf1``,
                ``FeedForwardSelf2``, ``FeedForwardSelf3``, ``FeedForwardSelf123``,
                ``BiFan``, ``CoRegulated``, ``CoRegulating``, ``BiParallel``,
                ``TriParallel``, ``Dominating4``, ``Dominating4Undir``, ``3Loop``,
                ``4Loop``, ``3LoopSelf123``, ``FourLoop``, ``FourCoLoop``,
                ``DirectedTwoLoop``, ``BiParallelLoop``, ``5Chain``, ``3Chain``,
                ``KeffStudy3``, ``KeffStudy4``, ``CoRegulatedSelf``, ``KeffLine4``,
                ``KeffLineLoop4``, ``3Full``, ``6Pyramid``, ``4Split``, ``5BiParallel``,
                ``6BiParallelDilation``, ``6BiParallelDilationLoop``, ``5combine``, ``4tree``.

    Returns:
        (networkx.DiGraph) : The directed graph motif.
    """

    graph = nx.DiGraph()

    if name == "FeedForward":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)

    elif name == "Fan":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)

    elif name == "FeedForwardSelf1":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(0, 0)

    elif name == "FeedForwardSelf2":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(1, 1)

    elif name == "FeedForwardSelf3":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(2, 2)

    elif name == "FeedForwardSelf123":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(0, 0)
        graph.add_edge(1, 1)
        graph.add_edge(2, 2)

    elif name == "BiFan":
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)

    elif name == "CoRegulated":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(2, 1)

    elif name == "CoRegulating":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 0)
        graph.add_edge(1, 2)

    elif name == "BiParallel":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)

    elif name == "TriParallel":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)
        graph.add_edge(0, 3)

    elif name == "Dominating4":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 1)

    elif name == "Dominating4Undir":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(1, 0)
        graph.add_edge(2, 0)
        graph.add_edge(3, 0)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 1)

    elif name == "3Loop":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 0)

    elif name == "4Loop":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 0)

    elif name == "3LoopSelf123":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 0)
        graph.add_edge(0, 0)
        graph.add_edge(1, 1)
        graph.add_edge(2, 2)

    elif name == "FourLoop":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 1)

    elif name == "FourCoLoop":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 1)

    elif name == "DirectedTwoLoop":
        graph.add_edge(0, 1)
        graph.add_edge(1, 0)
        graph.add_edge(2, 3)
        graph.add_edge(3, 2)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)

    elif name == "BiParallelLoop":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)
        graph.add_edge(3, 0)

    elif name == "5Chain":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)
        graph.add_edge(3, 4)

    elif name == "3Chain":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)

    elif name == "KeffStudy3":
        graph.add_edge(0, 1)
        graph.add_edge(1, 0)
        graph.add_edge(0, 2)
        graph.add_edge(2, 0)

    elif name == "KeffStudy4":
        graph.add_edge(0, 1)
        graph.add_edge(1, 0)
        graph.add_edge(0, 2)
        graph.add_edge(2, 0)
        graph.add_edge(0, 3)
        graph.add_edge(3, 0)

    elif name == "CoRegulatedSelf":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 2)
        graph.add_edge(2, 1)
        graph.add_edge(0, 0)

    elif name == "KeffLine4":
        graph.add_edge(0, 1)
        graph.add_edge(0, 3)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)

    elif name == "KeffLineLoop4":
        graph.add_edge(0, 1)
        graph.add_edge(0, 3)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)
        graph.add_edge(3, 0)

    elif name == "3Full":
        graph.add_edge(0, 0)
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 0)
        graph.add_edge(1, 1)
        graph.add_edge(1, 2)
        graph.add_edge(2, 0)
        graph.add_edge(2, 1)
        graph.add_edge(2, 2)

    elif name == "6Pyramid":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)
        graph.add_edge(1, 5)
        graph.add_edge(2, 3)
        graph.add_edge(2, 4)
        graph.add_edge(2, 5)

    elif name == "4Split":
        graph.add_edge(0, 1)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(2, 3)
        graph.add_edge(0, 3)

    elif name == "5BiParallel":
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(0, 4)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)

    elif name == "6BiParallelDilation":
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(0, 4)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)
        graph.add_edge(2, 5)
        graph.add_edge(3, 5)
        graph.add_edge(4, 5)

    elif name == "6BiParallelDilationLoop":
        graph.add_edge(0, 2)
        graph.add_edge(0, 3)
        graph.add_edge(0, 4)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)
        graph.add_edge(2, 5)
        graph.add_edge(3, 5)
        graph.add_edge(4, 5)
        graph.add_edge(5, 1)

    elif name == "5combine":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(0, 2)
        graph.add_edge(1, 3)
        graph.add_edge(3, 0)
        graph.add_edge(3, 4)
        graph.add_edge(4, 2)

    elif name == "4tree":
        graph.add_edge(0, 1)
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)

    else:
        raise TypeError('The motif name could not be found.')

    return graph



import random
from itertools import product, zip_longest


def recursive_map(f, d):
    """Normal python map, but recursive

    Args:
        f (function) : a function to be applied to every item of the iterable
        d (iterable) : the iterable to which f will be applied itemwise.
    """
    return [not hasattr(x, "__iter__") and f(x) or recursive_map(f, x) for x in d]


def binstate_to_statenum(binstate):
    """Converts from binary state to state number.

    Args:
        binstate (string) : The binary state.

    Returns:
        int : The state number.

    Example:

        .. code-block:: python

            '000' -> 0
            '001' -> 1
            '010' -> 2 ...

    See also:
        :attr:`statenum_to_binstate`, :attr:`statenum_to_density`
    """
    return int(binstate, 2)


def statenum_to_binstate(statenum, base):
    """Converts an interger into the binary string.

    Args:
        statenum (int) : The state number.
        base (int) : The binary base

    Returns:
        string : The binary state.

    Example:

        .. code-block:: python

            0 -> '00' (base 2)
            1 -> '01' (base 2)
            2 -> '10' (base 2)
            ...
            0 -> '000' (base 3)
            1 -> '001' (base 3)
            2 -> '010' (base 3)

    See also:
        :attr:`binstate_to_statenum`, :attr:`binstate_to_density`
    """

    return bin(statenum)[2:].zfill(base)


def binstate_pinned_to_binstate(binstate, pinned_binstate, pinned_var):
    """Combines two binstates based on the locations of pinned variables.

    Args:
        binstate (str) : the binary state of non-pinned variables
        pinned_binstate (str) : the binary states of the pinned variables
        pinned_var (list of int) : the list of pinned variables

    Returns:
        string : The combined binary state.

    See also:
        :attr:'statenum_to_binstate'
    """
    total_length = len(binstate) + len(pinned_binstate)
    new_binstate = list(statenum_to_binstate(0, base=total_length))
    ipin = 0
    ireg = 0
    for istate in range(total_length):
        if istate in pinned_var:
            new_binstate[pinned_var[ipin]] = pinned_binstate[ipin]
            ipin += 1
        else:
            new_binstate[istate] = binstate[ireg]
            ireg += 1
    return ''.join(new_binstate)


def statenum_to_output_list(statenum, base):
    """Converts an interger into a list of 0 and 1, thus can feed to BooleanNode.from_output_list()

    Args:
        statenum (int) : the state number
        base (int) : the length of output list

    Returns:
        list : a list of length base, consisting of 0 and 1

    See also:
        :attr:'statenum_to_binstate'
    """
    return [int(i) for i in statenum_to_binstate(statenum, base)]


def outputs_to_binstates_of_given_type(outputs, output, k):
    """Converts a node output list into a list of binary states given a specific output type.
    For instance, for the `AND` boolean function, with outputs `[0, 0, 0, 1]`
    it returns `['00', '01', '10']` when `output_type = 0` and `['11']` when `output_type = 1`.

    Args:
        outputs (list/str) : the list of outputs of a given node.
        output (int/str) : a certain output type. 0 or 1 in the case of Binary outputs.
        k (int) : input degree

    Returns:
        list : a list containing all the binstates of a certain output type

    """
    return [statenum_to_binstate(statenum, k) for statenum in range(2**k) if outputs[statenum] == output]


def flip_bit(bit):
    """Flips the binary value of a state.

    Args:
        bit (string/int/bool): The current bit position

    Returns:
        same as input: The flipped bit
    """
    if isinstance(bit, str):
        return '0' if (bit == '1') else '1'
    elif isinstance(bit, int) or isinstance(bit, bool):
        return 0 if (bit == 1) else 1
    else:
        raise TypeError("'bit' type format must be either 'string', 'int' or 'boolean'")


def flip_binstate_bit(binstate, idx):
    """Flips the binary value of a bit in a binary state.
        Args:
            binstate (string) : A string of binary states.
            idx (int) : The index of the bit to flip.
        Returns:
            (string) : New binary state.

        Example:

            .. code-block:: python
                flip_bit_in_strstates('000',1) -> '010'
        """
    if idx + 1 > len(binstate):
        raise TypeError("Binary state '{}' length and index position '{}' mismatch.".format(binstate, idx))
    return binstate[:idx] + flip_bit(binstate[idx]) + binstate[idx + 1:]


def statenum_to_density(statenum):
    """Converts from state number to density

    Args:
        statenum (int): The state number

    Returns:
        int: The density of ``1`` in that specific binary state number.

    Example:
        >>> statenum_to_binstate(14, base=2)
        >>> '1110'
        >>> statenum_to_density(14)
        >>> 3
    """
    return sum(map(int, bin(statenum)[2::]))


def binstate_to_density(binstate):
    """Converts from binary state to density

    Args:
        binstate (string) : The binary state

    Returns:
        int

    Example:
        >>> binstate_to_density('1110')
        >>> 3
    """
    return sum(map(int, binstate))


def binstate_to_constantbinstate(binstate, constant_template):
    """
    Todo:
        Documentation
    """
    # This function is being used in the boolean_network._update_trans_func
    constantbinstate = ''
    iorig = 0
    for value in constant_template:
        if value is None:
            constantbinstate += binstate[iorig]
            iorig += 1
        else:
            constantbinstate += str(value)

    return constantbinstate


def constantbinstate_to_statenum(constantbinstate, constant_template):
    """
    Todo:
        Documentation
    """
    binstate = ''.join([constantbinstate[ivar] for ivar in range(len(constant_template)) if constant_template[ivar] is None])
    return binstate_to_statenum(binstate)


def random_binstate(length, random_seed=None):
    """Generates a random binary state of a given length

    Args:
        length (int) : the length of the binary state

    Returns:
        binstate (str) : a random binary state
    """
    random.seed(random_seed)
    return "".join([random.choice(['0', '1']) for bit in range(length)])


def expand_logic_line(line):
    """This generator expands a logic line containing ``-`` (ie. ``00- 0`` or ``0-0 1``) to a series of logic lines containing only ``0`` and ``1``.

    Args:
        line (string) : The logic line. Format is <binary-state><space><output>.

    Returns:
        generator : a series of logic lines

    Example:
        >>> expand_logic_line('1-- 0')
        >>> 100 0
        >>> 101 0
        >>> 110 0
        >>> 111 0
    """
    # helper function for expand_logic_line
    def _insert_char(la, lb):
        lc = []
        for i in range(len(lb)):
            lc.append(la[i])
            lc.append(lb[i])
        lc.append(la[-1])
        return ''.join(lc)

    line1, line2 = line.split()
    chunks = line1.split('-')
    if len(chunks) > 1:
        for i in product(*[('0', '1')] * (len(chunks) - 1)):
            yield _insert_char(chunks, i) + ' ' + line2
    else:
        for i in [line]:
            yield i


def binstate_compare(binstate1, binstate2):
    """Compare each element in two binary states

    Args:
        binstate1, binstate2 : the two binary states to be compared

    Return:
        c (list, bool) : a list of comparisons
    """
    return [(b0 == b1) for b0, b1 in zip_longest(binstate1, binstate2)]


def hamming_distance(s1, s2):
    """Calculates the hamming distance between two configurations strings.

    Args:
        s1 (string): First string
        s2 (string): Second string

    Returns:
        float : The Hamming distance

    Example:
        >>> hamming_distance('001','101')
        >>> 1
    """
    assert len(s1) == len(s2), "The two strings must have the same length"
    return sum([(b0 != b1) for b0, b1 in zip_longest(s1, s2)])



# -*- coding: utf-8 -*-
"""
(Cythonized) Boolean Canalization
=====================

Functions to compute the Quine-McCluskey algorithm in cython for increaed computation speed.

"""
#   Copyright (C) 2021 by
#   Rion Brattig Correia <rionbr@gmail.com>
#   Alex Gates <ajgates42@gmail.com>
#
#   All rights reserved.
#   MIT license.
from cana.cutils import *

WILDCARD_SYMBOL = '#'
SYMMETRIC_WILDCARD_SYMBOL = '*'


#
# Quine-McCluskey Functions
#
def make_density_groups(input_binstates):
    """
    """

    density_groups = dict()
    for binstate in input_binstates:
        density = binary_density(binstate)
        if density not in density_groups:
            density_groups[density] = set()
        density_groups[density].add(binstate)

    return density_groups


def find_wildcards(binstate1, binstate2):
    """
    Compare two binary states and replace any differing bits by a wildcard.
    Args:
        binstate1, binstate2 : the two binary states to be compared

    Return:
        c (list, bool) : a list of comparisons

    """
    # assert len(s1) == len(s2) , "The two binstates must have the same length"
    return "".join([b0 if (b0 == b1) else WILDCARD_SYMBOL for b0, b1 in zip(binstate1, binstate2)])


def binary_density(binstate):
    """
    Find the density (number of 1s) for a term with possible wildcards.
    """
    return binstate.count('1')


def replace_wildcard(binstate, idx):
    """
    Return the binary state with a wildcard at the idx position.
    """
    return binstate[:idx] + WILDCARD_SYMBOL + binstate[idx + 1:]


def find_implicants_qm(input_binstates, verbose=False):
    """ Finds the prime implicants (PI) using the Quine-McCluskey algorithm :cite:`Quine:1955`.

    Args:
        input_binstates (list / set) : A the binstates to condense.

    Returns:
        PI (set): a set of prime implicants.

    # Authors: Alex Gates
    """

    # we start with an empty set of implicants
    matched_implicants = set()
    done = False

    # repeat the following until no matches are found
    while not done:

        # split up the input_binstates into groups based on the number of 1s (density)
        density_groups = make_density_groups(input_binstates)

        # now clear everything for the new pass
        input_binstates = set()
        used = set()

        # Find the prime implicants

        # for each possible density
        for density in density_groups:
            # first make sure there are other binstates with the next density
            if density + 1 in density_groups:

                # then we pass through the binstates
                for binstate0 in density_groups[density]:

                    # An optimization due to Thomas Pircher, https://github.com/tpircher/quine-mccluskey/blob/master/quine_mccluskey/qm.py
                    # The Quine-McCluskey algorithm compares t1 with
                    # each element of the next group. (Normal approach)
                    # But in reality it is faster to construct all
                    # possible permutations of t1 by adding a '1' in
                    # opportune positions and check if this new term is
                    # contained in the set groups[key_next].

                    for idx, b0 in enumerate(binstate0):
                        if b0 == '0':
                            binstate1 = flip_binstate_bit(binstate0, idx)
                            if binstate1 in density_groups[density + 1]:
                                # keep track of the covered binary states
                                used.add(binstate0)
                                used.add(binstate1)
                                # keep the new wildcard binstate for the next round
                                input_binstates.add(replace_wildcard(binstate0, idx))

        # now add back the implicants that were not matched
        for groups in list(density_groups.values()):
            matched_implicants |= groups - used

        # finally, check if this pass actually compressed any terms
        # we finish when we cant make any further compressions
        if len(used) == 0:
            done = True

    # finish up by adding back all of the uncovered binary states
    prime_implicants = matched_implicants
    for groups in list(density_groups.values()):
        prime_implicants |= groups

    return prime_implicants


def __pi_covers(implicant, binstate):
    """Determines if a binarystate is covered by a specific implicant.
    Args:
        implicant (string): the implicant.
        minterm (string): the minterm.
    Returns:
        x (bool): True if covered else False.

    """
    return all(i == WILDCARD_SYMBOL or m == i for i, m in zip(implicant, input))


def expand_wildcard_schemata(schemata):
    """
    Expand a wildcard schemata to list all binary states it covers.

    Args:
        schemata (string): the wildcard shemata
    Returns:
        binary_states (list): list of all binary states covered by the schemata
    """

    # count the number of wildcard symbols
    nwildcards = schemata.count(WILDCARD_SYMBOL)

    # if there arent any symbols, return the original schemata
    if nwildcards == 0:
        return [schemata]
    else:
        binary_states = []
        for wildstatenum in range(2**nwildcards):
            wildstates = statenum_to_binstate(wildstatenum, nwildcards)
            wnum = 0
            newstate = ''
            for b in schemata:
                if b == WILDCARD_SYMBOL:
                    newstate += wildstates[wnum]
                    wnum += 1
                else:
                    newstate += b
            binary_states.append(newstate)
        return binary_states


def return_pi_coverage(prime_implicants):
    """Computes the binary states coverage by Prime Implicant schematas.

    Args:
        prime_implicants (set): a set of prime implicants.
            This is returned by `find_implicants_qm`.
    Returns:
        pi_coverage (dict) : a dictionary of coverage where keys are input states and values are lists of the Prime Implicants covering that input.
    """

    pi_coverage = dict()
    for pi in prime_implicants:
        for binstate in expand_wildcard_schemata(pi):
            if binstate not in pi_coverage:
                pi_coverage[binstate] = set()
            pi_coverage[binstate].add(pi)

    return pi_coverage


def input_wildcard_coverage(pi_coverage):
    """Computes the binary states coverage by Prime Implicant schematas.

    Args:
        pi_coverage (dict): a dict mapping binary states to their prime implicants.
            This is returned by `return_pi_coverage`.
    Returns:
        input_wildcard_coverage (dict) : a dictionary of coverage where keys are inputs and values are lists of wither a WildCard covers that input.

    """
    # number of inputs
    k = len(next(iter(pi_coverage)))

    input_to_wildcards = {i: dict() for i in range(k)}
    for binstate, piset in pi_coverage.items():
        for i in range(k):
            input_to_wildcards[i][binstate] = tuple(pi[i] == WILDCARD_SYMBOL for pi in piset)

    return input_to_wildcards



# -*- coding: utf-8 -*-
"""
BNS Attractors Interface
==========================

This module interfaces CANA with the [B]oolean [N]etworks with [S]ynchronous update (BNS) :cite:`Dubrova:2011` to compute attractors.

    BNS is a software tool for computing attractors in Boolean Networks with Synchronous update.
    Synchronous Boolean networks are used for the modeling of genetic regulatory networks.

    BNS implements the algorithm presented in which is based on a SAT-based bounded model checking.
    BNS uses much less space compared to BooleNet or other BDD-based approaches for computing attractors.
    It can handle several orders of magnitude larger networks.



.. Note::

    You must have ``bns`` compiled for you system. Alternatively, you can download the binary from the `bns website <https://people.kth.se/~dubrova/bns.html>`_ directly.
    Last updated: March 2017.

"""
#   Copyright (C) 2017 by
#   Rion Brattig Correia <rionbr@gmail.com>
#   Alex Gates <ajgates@indiana.edu>
#   All rights reserved.
#   MIT license.
import os
import subprocess
import tempfile
from cana.utils import binstate_to_statenum

_path = os.path.dirname(os.path.realpath(__file__))
""" Make sure we know what the current directory is """


def attractors(cnet, bnspath=_path, cleanup=True):
    """Makes a subprocess call to `bns` supplying a temporary file with the boolean logic.

    Args:
        cnet (file,string) : A .cnet formated string or file.
        bnspath (string) : The path to the bns binary.
        cleanup (bool) : If cnet is a string, this function creates a temporary file. This forces the removal of this temp file.

    Returns:
        list : the list of attractors
    """

    # If is file, open the file
    if os.path.isfile(cnet):
        file = cnet

    # If string, Creates a Temporary File to be supplied to BNS
    elif isinstance(cnet, str):
        tmp = tempfile.NamedTemporaryFile(delete=cleanup)
        with open(tmp.name, 'w') as openfile:
            openfile.write(cnet)
        tmp.file.close()
        file = tmp.name
    else:
        raise TypeError('The cnet input should be either a file to a .cnet file or a string containing the .cnet content')

    cmd = [os.path.join(_path, 'bns'), file]
    attractors = list()

    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        current_attractor = []
        for i, line in enumerate(p.stdout):
            # Strip line
            cleanline = line.decode('ascii').strip().replace('\n', '')

            if 'Attractor' in cleanline:
                attractors.append(current_attractor)
                current_attractor = []
            elif 'Node' in cleanline and 'assumed to be constant' in cleanline:
                pass
            elif 'Total' in cleanline:
                pass
            elif 'Start searching for all atractors.' in cleanline:
                pass
            elif 'Depth' in cleanline:
                pass
            elif 'average' in cleanline:
                pass
            elif len(cleanline) > 0:
                current_attractor.append(binstate_to_statenum(cleanline))

    except OSError:
        print("'BNS' could not be found! You must have it compiled or download the binary for your system from the 'bns' website (https://people.kth.se/~dubrova/bns.html).")

    return attractors



# -*- coding: utf-8 -*-
"""
Boolean Canalization
=====================

Functions to compute the Quine-McCluskey algorithm.

"""
#   Copyright (C) 2021 by
#   Rion Brattig Correia <rionbr@gmail.com>
#   Alex Gates <ajgates@gmail.com>
#   Etienne Nzabarushimana <enzabaru@indiana.edu>
#   All rights reserved.
#   MIT license.

import numpy as np

import itertools
from .. utils import *

from collections import deque

__author__ = """\n""".join([
    'Alex Gates <ajgates@umail.iu.edu>',
    'Etienne Nzabarushimana <enzabaru@indiana.edu>',
    'Rion Brattig Correia <rionbr@gmail.com>'
])


# Quine-McCluskey Functions
def make_transition_density_tables(k=1, outputs=[0, 1]):
    """ This method creates a tuple-of-lists that is used to calculate Prime Implicants in the first step of the Quine-McCluskey algorithm :cite:`Quine:1955`.
    In practice it separates the positive and negative transitions (tuple), then further separates it by counting the number of 1's in each (lists).

    Args:
        k (int) : the ``k`` number of inputs
        outputs (list) : a list of ``[0,1]`` output for each state number.

    Returns:
        tables (tuple) : a tuple where [0] is the negative table and [1] is the positive table.
    """
    # make sure outputs are integers
    outputs = list(map(int, outputs))

    # we need to split up the LUT based on the transition (to either 0 or 1) and the density of 1s in the binstate
    transition_density_tuple = [[[] for density in range(k + 1)] for transition in [0, 1]]
    for statenum in range(2**k):
        binstate = statenum_to_binstate(statenum, base=k)
        density = binstate_to_density(binstate)
        transition = outputs[statenum]
        # Account for Dont-Care (2) transition states
        if transition == 2:
            transition_density_tuple[0][density].append(binstate)
            transition_density_tuple[1][density].append(binstate)
        else:
            transition_density_tuple[transition][density].append(binstate)
    #
    return transition_density_tuple


def find_implicants_qmOLD(column, verbose=False):
    """ Finds the prime implicants (PI) using the Quine-McCluskey algorithm :cite:`Quine:1955`.

    Args:
        column (list) : A list-of-lists containing the counts of ``1`` for each input.
            This is given by `make_transition_density_tables`.

    Returns:
        PI (set): a set of prime implicants.

    # Authors: Alex Gates and Etienne Nzabarushimana
    """

    N = len(column) - 1

    # we start with an empty set of implicants
    prime_implicants = set()
    done = False

    # repeat the following until no matches are found
    while not done:
        done = True

        # default everything to empty with no matches
        next_column = [set() for _ in range(N + 1)]
        matches = [[False for _ in range(len(column[density]))] for density in range(N + 1)]

        # loop through the possible densities
        for density in range(N):

            # compare the implicants from successive densities
            for i, implicant in enumerate(column[density]):
                for j, candidate in enumerate(column[density + 1]):

                    # check if the implicants differ on only one variable
                    match = _adjacent(implicant, candidate)
                    if match:
                        matches[density][i] = matches[density + 1][j] = True
                        matches_density = sum([var != '0' for var in match])
                        next_column[matches_density].add(match)
                        done = False

        # now add back the implicants that were not matched
        for i in range(N + 1):
            for j in range(len(matches[i])):
                if not matches[i][j]:
                    prime_implicants.add(column[i][j])

        # use the simplified table as the starting point of the next pass
        column = [list(g) for g in next_column]

    return prime_implicants


def _adjacent(imp1, imp2):
    """Determine if two implicants are adjacent: ie differ on only one variable.

    Args:
        imp1 (string): implicant 1
        imp1 (string): implicant 2

    Returns:
        (bool)
    """
    differences = 0
    match = []
    for m1, m2 in zip(imp1, imp2):
        if m1 == m2:
            match.append(m1)
        elif differences:
            return False
        else:
            differences += 1
            match.append('2')
    return "".join(match)


def __pi_covers(implicant, input, symbol=['2', '#', 2]):
    """Determines if a minterm is covered by a specific implicant.

    Args:
        implicant (string): the implicant.
        minterm (string): the minterm.

    Returns:
        x (bool): True if covered else False.

    """
    for i, m in zip(implicant, input):
        if i in symbol:
            continue
        if int(i) != int(m):
            return False
    return True


def computes_pi_coverage(k, outputs, prime_implicants):
    """Computes the input coverage by Prime Implicant schematas.

    Args:
        k (int): the number of inputs.
        outpus (list): the list of transition outputs.
        prime_implicants (tuple): a tuple containing a list negative and positive prime implicants. This is returned by `find_implicants_qm`.

    Returns:
        pi_coverage (dict) : a dictionary of coverage where keys are input states and values are lists of the Prime Implicants covering that input.

    Note: based on code from Alex Gates and Etienne Nzabarushimana.
    """
    # make sure outputs are integers
    outputs = list(map(int, outputs))

    pi_coverage = {}
    for statenum in range(2**k):
        binstate = statenum_to_binstate(statenum, base=k)
        pi_coverage[binstate] = covering_implicants = []
        transition = outputs[statenum]
        # Add support for DontCare (2) transition
        if transition == 2:
            transition = [0, 1]
        else:
            transition = [outputs[statenum]]
        for t in transition:
            for prime_implicant in prime_implicants[t]:
                if __pi_covers(prime_implicant, binstate):
                    covering_implicants.append(prime_implicant)
    #
    return pi_coverage


# Two Symbols Functions
def find_two_symbols_v2(k=1, prime_implicants=None, verbose=False, verbose_level=0):
    """This function calculates the permutation, two-symbol (TS), list of schematas.
    This implementation considers '11' and '00' as a possible input permutation.

    Args:
        k (int): The number of inputs.
        prime_implicants (list): The prime implicants computed.

    Returns:
        final_list (list) : The list of two-symbol schematas.

    Note: This is a modification of the original algorithm that can be found in Marques-Pita & Rocha [2013].
    """
    if not len(prime_implicants):
        return []

    # If this node has no input, yet it affects other nodes (fixed variable)
    if k == 0:
        TSf = []
        for pi in prime_implicants:
            TSf.append((pi, [], []))
        return TSf

    # Init
    # n_pi = len(prime_implicants) # never used
    pi_matrix = np.array(tuple(map(tuple, prime_implicants)), dtype=int)

    # List of the Two-Symbol Schematas
    TS = []

    # Init Queue
    Q = deque()
    Q_history = set()
    Q.append(pi_matrix)
    i = 0

    while len(Q):

        schematas = Q.popleft()
        n_schematas = schematas.shape[0]
        i += 1

        if verbose:
            if verbose_level == 1:
                if (i % 500 == 0):
                    print('>>> QUEUE: pop | A (m=%d) | Queue size: %d' % (n_schematas, len(Q)))
            elif verbose_level > 5:
                print('>>> QUEUE: pop | A (m=%d) | Queue size: %d' % (n_schematas, len(Q)))
            if verbose_level > 10:
                print(schematas)

        # count the number of [0's, 1's, 2's] in each column
        column_counts = _count_cols_symbols_v2(pi_matrix=schematas, verbose=verbose, verbose_level=verbose_level)
        if verbose and verbose_level > 10:
            print('>>> COLUMN Schema Counts:')
        # find the permutation groups based on column counts
        perm_groups = _check_col_counts_v3(counts_matrix=column_counts, verbose=verbose, verbose_level=verbose_level)

        if (perm_groups != -1):
            if verbose and verbose_level > 10:
                print('>>> There are permutable groups! Lets loop them')
            for x_group in perm_groups:
                if verbose and verbose_level > 20:
                    print('>>> x_group:', x_group)
                    print('>>> Truncated schemata matrix:')
                    print(schematas[:, x_group].T)
                # find the row counts by taking the transpose of the truncated schemata list
                row_counts = _count_cols_symbols_v2(pi_matrix=schematas[:, x_group].T, verbose=verbose, verbose_level=verbose_level)
                if verbose and verbose_level > 20:
                    print('>>> ROW Schema Counts:')
                    print(row_counts)
                # make sure all row counts are the same
                if not (row_counts == row_counts[0]).all():
                    if verbose and verbose_level > 20:
                        print('>>> row_counts are NOT the same (-1)')
                    perm_groups = -1

        if verbose and verbose_level > 10:
            print(">>> Exists permutation groups?:", (perm_groups != -1))
            print(">>> Are groups already in F''?:", ((schematas.tolist(), perm_groups) in TS))
        if (perm_groups != -1) and not ((schematas.tolist(), perm_groups) in TS):
            # do some weird permutation group testing
            allowed_perm_groups = _check_schemata_permutations_v2(schematas, perm_groups, verbose=verbose, verbose_level=verbose_level)
            if verbose and verbose_level > 15:
                print('>>> Permutation testing result:', allowed_perm_groups)
            if allowed_perm_groups is not None:
                if verbose and verbose_level > 15:
                    print(">>> RESULTS: adding F'': %s , Idxs: %s" % (schematas.tolist(), allowed_perm_groups))
                TS.append((schematas.tolist(), allowed_perm_groups))

        else:
            if verbose and verbose_level > 10:
                print('>>> Generate combinations of schematas (m-1) and add to Queue')
            if schematas.shape[0] > 2:
                for idxs_subset in itertools.combinations(np.arange(0, n_schematas), (n_schematas - 1)):
                    idxs_subset = list(idxs_subset)
                    schemata_subset = schematas[idxs_subset, :]
                    # This schemata has already been inserted onto the Queue before?
                    if schemata_subset.tostring() not in Q_history:
                        if verbose and verbose_level > 25:
                            print('>>> QUEUE: appending (idxs: %s)' % (idxs_subset))
                            print(schemata_subset)
                        Q.append(schemata_subset)
                        Q_history.add(schemata_subset.tostring())
                    else:
                        if verbose and verbose_level > 25:
                            print('>>> QUEUE: duplicate, skip (idxs: %s)' % (idxs_subset))

    if verbose:
        print('>>> TWO-SYMBOLS:')
        for i, (tss, perms) in enumerate(TS):
            print("F''-%d: %s | Perms: %s" % (i, tss, perms))

    # Simplification. Check if there are TSs that are completely contained within others.
    # 'ts' = Two-Symbol
    # 'cx' = Complexity
    # 'xs' = Expanded Logic
    TSs = {
        i: {
            'tss': tss,
            'perms': perms,
            'cx': _calc_ts_complexity(tss, perms),
            'xl': _expand_ts_logic(tss, perms)
        } for i, (tss, perms) in enumerate(TS)
    }

    # Loops all combinations (2) of TS
    for (i, j) in itertools.combinations(TSs.keys(), 2):
        try:
            a_in_b, b_in_a = _check_schema_within_schema(TSs[i]['xl'], TSs[j]['xl'])
        except:
            continue
        else:
            cx_a = TSs[i]['cx']
            cx_b = TSs[j]['cx']
            # A or B contained in the other, keep only contained.
            if a_in_b and not b_in_a:
                del TSs[i]
            elif b_in_a and not a_in_b:
                del TSs[j]
            elif a_in_b and b_in_a:
                # Keep most complex
                if cx_a < cx_b:
                    del TSs[i]
                elif cx_b < cx_a:
                    del TSs[j]
                else:
                    # they are equal, delete either one
                    del TSs[i]

    if verbose:
        print('>>> TWO-SYMBOLS (simplified):')
        for i, (tss) in TSs.items():
            print("F''-%d: %s | Perms: %s | CX: %s" % (i, tss['tss'], tss['perms'], tss['cx']))

    # Final List (from simplified)
    TSf = [(tss['tss'][0], tss['perms'], []) for tss in TSs.values()]

    # Check if all PI are being covered. If not, include the PI on the TS list
    if verbose:
        print('>>> Check all PI are accounted for in the TS')
    for i, pi in enumerate(pi_matrix, start=0):
        if not any([_check_schema_within_schema([pi.tolist()], tss['xl'], dir='a', verbose=verbose)[0] for tss in TSs.values()]):
            if verbose:
                print("PI-%d '%s' Not in list, ADDING." % (i, pi.tolist()))
            TSf.append((pi.tolist(), [], []))
        else:
            if verbose:
                print("PI-%d '%s' OK." % (i, pi.tolist()))

    if verbose:
        print('>>> Check for Same-Symbol permutables')

    # NEW: Step to include same-symbol permutables
    for ts, perms, sames in TSf:
        # Indices of permutables inputs
        idxs = list(set([idx for idxs in perms for idx in idxs]))

        # Makes the F'' into a Collum Array so it can be used by '_count_cols_symbols_vX'
        ts_matrix = np.array([ts]).T

        # Remove Inputs (columns) that already have permutable symbols. Only if there are permutables
        if len(idxs):
            rmask = np.array(idxs)
            ts_matrix_left = ts_matrix[~rmask, :]
        else:
            ts_matrix_left = ts_matrix

        if verbose and verbose_level > 10:
            print("> F'' Original:")
            print(ts_matrix)
            print("> Permutables: %s" % (perms))
            print("> F'' without permutables:")
            print(ts_matrix_left)

        counts_matrix = _count_cols_symbols_v2(pi_matrix=ts_matrix_left.T, verbose=False, verbose_level=verbose_level)
        perm_groups = _check_identical_cols_count_symbols_v2(counts_matrix=counts_matrix, verbose=verbose, verbose_level=verbose_level)
        sames.extend(perm_groups)

    # Step to convert the pi list to string
    for i, (ts, perms, sames) in enumerate(TSf, start=0):
        ts = ''.join(map(str, ts))
        TSf[i] = (ts, perms, sames)

    # Final list after all PI were accounted for
    if verbose:
        print('>>> TS (final list):')
        for i, tss, sms in TSf:
            print("TS: '%s' | Perm Idx: %s | Sms Idx: %s" % (i, tss, sms))

    return TSf


def _calc_ts_complexity(tss, pers):
    """ Calculates the complexity of a TS schema
    Complexity = (Number of Schemas + Number of Permutable Symbols + Lenght of each Permutable Symbol)
    """
    return len(tss) + sum([len(per) for ts, per in zip(tss, pers)])


def _check_schema_within_schema(la, lb, dir=None, verbose=False):
    """ Check is a Two-Symbol schemata is covered by another.
    This is used to simplify the number of TS schematas returned.
    The arguments for this function are generated by `_expand_ts_logic`.

    Args:
        tsa (list) : A list of :math:`F'` schematas that a Two-Symbol :math:`F''` schemata can cover.
        tsb (list) : A list of :math:`F'` schematas that a Two-Symbol :math:`F''` schemata can cover.
        dir (string) : The direction to check, either ``a`` or ``b`` is in the other.
            Defaults to both directions.
    """
    a_in_b, b_in_a = None, None
    #
    if dir != 'b':
        a_in_b = all([(xa in lb) for xa in la])
        if verbose:
            print('%s in %s : %s' % (la, lb, a_in_b))
    if dir != 'a':
        b_in_a = all([(xb in la) for xb in lb])
        if verbose:
            print('%s in %s : %s' % (lb, la, b_in_a))
    #
    return a_in_b, b_in_a


def _expand_ts_logic(two_symbols, permut_indexes):
    """ Expands the Two-Symbol logic to all possible prime-implicants variations being covered.

    Args:
        two_symbols (list) : Two-Symbol schematas list-of-lists.

    Returns:
        (list) : a list of :math:`F'` covered by this Two-Symbol.
    """
    # If receiving a binary string, convert to list of lists
    if isinstance(two_symbols, str):
        two_symbols = [list(two_symbols)]
    # Queue
    Q = deque()
    Q.extend(two_symbols)
    logics = []
    #
    while Q:
        implicant = np.array(Q.pop())
        for idxs in permut_indexes:
            # Permutation of all possible combinations of the values that are permutable.
            for vals in itertools.permutations(implicant[idxs], len(idxs)):
                # Generate a new schema
                _implicant = copy.copy(implicant)
                _implicant[idxs] = vals
                # Insert to list of logics if not already there
                if not(_implicant.tolist() in logics):
                    logics.append(_implicant.tolist())
                    Q.append(_implicant.tolist())
    return logics


def _check_schemata_permutations_v2(schematas, perm_groups, verbose=False, verbose_level=0):
    """ Checks if the permutations are possible

    Note:
        Not sure if this is really needed.
    """
    if verbose and verbose_level > 20:
        print("-- Check Schemata Permutations (v2) : g(H',L) --")
    allowed_perm_groups = []
    all_indices = set([i_index for x_group in perm_groups for i_index in x_group])
    for x_group in perm_groups:
        sofar = []
        for i_index in range(len(x_group) - 1):
            x_index = x_group[i_index]
            small_group = [x_index]
            if not (x_index in sofar):
                sofar.append(x_index)

                for y_index in x_group[(i_index + 1)::]:

                    if (not(y_index in sofar)) and _can_swap_v2(schematas[:, [x_index, y_index]], verbose=verbose, verbose_level=verbose_level):
                        small_group.append(y_index)
                        sofar.append(y_index)
                if len(small_group) > 1:
                    allowed_perm_groups.append(small_group)

        if verbose and verbose_level > 30:
            print('> allowed_perm_groups', allowed_perm_groups)
        if set([i_index for x_group in allowed_perm_groups for i_index in x_group]) == all_indices:
            return allowed_perm_groups
    return None


def _can_swap_v2(schemata_subset, verbose=False, verbose_level=0):
    """Determines if two schemata subsets can be swapped"""
    if verbose and verbose_level > 40:
        print('> Can Swap?:',)
    can_switch = 1
    for row in schemata_subset[:, [1, 0]]:
        can_switch *= np.any(np.all(schemata_subset == row, axis=1))
    if verbose and verbose_level > 40:
        print(can_switch)
    return can_switch


def _check_col_counts_v3(counts_matrix, verbose=False, verbose_level=0):
    """ This function is used to find permutable symbols.

    Args:
        counts_matrix (numpy.ndarray) : a matrix where rows are inputs and columns are possible input types (0,1 or #)

    Returns:
        perm_groups (list) : a list of the indexes that can be permuted.
    """
    if verbose and verbose_level > 30:
        print('-- Check Col Counts (v3) --')

    counts = {}         # Multi Counts
    perm_groups = []    # A list of groups of Permutable Indexes

    for i, row in enumerate(counts_matrix, start=0):
        # a tuple (hashable) version of the row counts
        row_tuple = tuple(row)

        if row_tuple in counts:
            # we have seen this one before, so add it to the permutation group
            counts[row_tuple].append(i)
        elif np.count_nonzero(row) >= 2:
            # we have not seen this count before, it is not a fixed variable, so create a new entry for it
            counts[row_tuple] = [i]
        else:
            # we will skip fixed variables
            pass

    # Append non-constants that have permutable positions
    for col, idxs in counts.items():
        if verbose and verbose_level > 40:
            print(col, ':', idxs)

        if len(idxs) == 1:
            return -1
        elif len(idxs) >= 1:
            perm_groups.append(idxs)

    if verbose and verbose_level > 40:
        print('counts:', counts)
        print('perm_groups:', perm_groups)

    if len(perm_groups):
        return perm_groups
    else:
        return -1


def _check_identical_cols_count_symbols_v2(counts_matrix, verbose=False, verbose_level=0):
    """ This function is used to find same symbol permutables. In practice it is a variance of `_check_cols_symbols_vX`

    Args:
        counts_matrix (numpy.ndarray) : a matrix where rows are inputs and columns are possible input types (0,1 or #)

    Returns:
        perm_groups (list) : a list of the indexes that can be permuted
    """
    if verbose and verbose_level > 20:
        print('-- Check Identical Col Counts (v2) --')

    counts = {}         # Multi Counts
    perm_groups = []    # A list of groups of Permutable Indexes

    for i, row in enumerate(counts_matrix, start=0):
        # a tuple (hashable) version of the row counts
        row = row.tolist()
        row_tuple = tuple(row)

        if verbose and verbose_level > 30:
            print('RC: %s : %s' % (i, row_tuple))

        if row_tuple in counts:
            # we have seen this one before, so add it to the permutation group
            counts[row_tuple].append(i)
        else:
            # we have not seen this count before, so create a new entry for it
            counts[row_tuple] = [i]

    # Append non-constants that have permutable positions
    for col, idxs in counts.items():
        if verbose and verbose_level > 30:
            print(col, ':', idxs)

        if len(idxs) >= 2:
            perm_groups.append(idxs)

    if verbose and verbose_level > 30:
        print('counts:', counts)
        print('sames_groups:', perm_groups)

    if len(perm_groups):
        return perm_groups
    else:
        return []


def _count_cols_symbols_v2(pi_matrix=None, verbose=False, verbose_level=0):
    """ Given a matrix, where each row is a prime implicant, counts how many 0's, 1's and 2's are found in each column.

    Args:
        pi_matrix (numpy.ndarray) : a matrix ``n \times k`` of ``n`` prime implicants.

    Returns:
        counts (numpy.ndarray) : a matrix ``n \times 3`` where the entries are counts.
    """
    if verbose and verbose_level > 20:
        print(' -- Count Cols (v2) --')
    # How many PI?
    n = pi_matrix.shape[1]
    # Instanciate count matrix
    counts = np.zeros((n, 3), dtype=int)
    for i, col in enumerate(pi_matrix.T):
        # Count how many values are found and update the matrix of counts
        val, cnt = np.unique(col, return_counts=True)
        counts[i, val] = cnt

    return counts


############ START OF TWO SYMBOL v.1 ############


# This version does not conside '11' and '00' as permutable symbols and had other bugs solved by v2
def find_two_symbols_v1(k=1, prime_implicants=None, verbose=False):
    two_symbol_schemata_list = []

    Partition_Options = [prime_implicants]
    while len(Partition_Options) > 0:
        # take first partition out of the set
        schemata_list = np.array(map(list, Partition_Options.pop()))
        if verbose:
            print('== Partitions (v1) ==')
            print('>>> A (m=%d)' % schemata_list.shape[0])
            print(schemata_list)

        # count the number of [0's, 1's, 2's] in each column
        column_counts = _three_symbol_count_cols_symbols_v1(k=k, transition_list=schemata_list)

        # find the permutation groups based on column counts
        permutation_groups = _check_counts_v1(column_counts)

        if (permutation_groups != -1):
            if verbose:
                print('>>> There are permutable groups! Lets loop them')
            for x_group in permutation_groups:

                # find the row counts by taking the transpose of the truncated schemata list
                row_counts = _three_symbol_count_cols_symbols_v1(k=schemata_list.shape[0], transition_list=schemata_list[:, x_group].T)
                if verbose:
                    print('>>> ROW Schema Counts:')
                    print(row_counts)
                # make sure all row counts are the same
                if len(row_counts) != row_counts.count(row_counts[0]):
                    permutation_groups = -1
        if verbose:
            print(">>> Permutation groups:", (permutation_groups != -1), permutation_groups)
            print(">>> Permuted groups already in F'':", ((schemata_list.tolist(), permutation_groups) in two_symbol_schemata_list))
        if (permutation_groups != -1) and not ((schemata_list.tolist(), permutation_groups) in two_symbol_schemata_list):
            # do some weird permutation group testing
            allowed_permutation_groups = _check_schemata_permutations_v1(schemata_list, permutation_groups)
            if allowed_permutation_groups != []:
                if verbose:
                    print("ADDING to F'':", schemata_subset)
                two_symbol_schemata_list.append((schemata_list.tolist(), allowed_permutation_groups))

        else:
            if schemata_list.shape[0] > 2:
                for schemata_subset in itertools.combinations(schemata_list, (schemata_list.shape[0] - 1)):
                    if verbose:
                        print('ADDING to Queue:', schemata_subset)
                    Partition_Options.append(np.array(schemata_subset))

    if verbose:
        print('Partition_Options:', Partition_Options)
    final_list = []
    prime_accounted = []

    for p_implicant in prime_implicants:
        p_implicant = list(p_implicant)
        if not (p_implicant in prime_accounted):
            for f_double_prime, r_perm in two_symbol_schemata_list:
                for f_prime in f_double_prime:
                    if np.all(f_prime == p_implicant):
                        final_list.append((p_implicant, r_perm))
                        for account_prime in f_double_prime:
                            prime_accounted.append(list(account_prime))
        if not (p_implicant in prime_accounted):
            final_list.append((p_implicant, []))

    return final_list

def _check_schemata_permutations_v1(schemata_list, permutation_groups):

    allowed_permutation_groups = []
    all_indices = set([i_index for x_group in permutation_groups for i_index in x_group])

    for x_group in permutation_groups:
        sofar = []

        for i_index in range(len(x_group) - 1):
            x_index = x_group[i_index]
            small_group = [x_index]

            if not (x_index in sofar):
                sofar.append(x_index)

                for y_index in x_group[(i_index + 1)::]:
                    if (not(y_index in sofar)) and _can_swap_v1(schemata_list[:, [x_index, y_index]]):
                        small_group.append(y_index)
                        sofar.append(y_index)
                if len(small_group) > 1:
                    allowed_permutation_groups.append(small_group)

        if set([i_index for x_group in allowed_permutation_groups for i_index in x_group]) == all_indices:
            return allowed_permutation_groups
    return []


def _can_swap_v1(schemata_subset):
    can_switch = 1
    for row in schemata_subset[:, [1, 0]]:
        can_switch *= np.any(np.all(schemata_subset == row, axis=1))
    return can_switch


def _check_counts_v1(column_counts=[]):
    print('-- Column Counts (v1) --')
    print(column_counts)
    unique_col_counts = []
    permutation_groups = []

    for i_col_count, x_col_count in enumerate(column_counts):
        print('RC: %d : %s' % (i_col_count, x_col_count))
        if x_col_count.count(0) >= 2:
            # this is a constant column so skip it
            pass

        elif x_col_count in unique_col_counts:
            # we have seen this one before, so add it to the permutation group
            permutation_groups[unique_col_counts.index(x_col_count)].append(i_col_count)

        else:
            # we have not seen this count before, so create a new entry for it
            unique_col_counts.append(x_col_count)
            permutation_groups.append([i_col_count])

    # check if a singleton permutation group exists
    if [len(x_group) for x_group in permutation_groups].count(1) > 0:
        print('counts:', permutation_groups)
        return -1
    else:
        print('counts:', permutation_groups)
        return permutation_groups


def _three_symbol_count_cols_symbols_v1(k=1, transition_list=None):
    column_counts = [[0, 0, 0] for i_col in range(k)]
    for x_col in transition_list:
        for i_entry, x_entry in enumerate(x_col):
            if x_entry == '0':
                column_counts[i_entry][0] += 1
            elif x_entry == '1':
                column_counts[i_entry][1] += 1
            elif x_entry == '2':
                column_counts[i_entry][2] += 1

    return column_counts


############ END OF TWO SYMBOL v.1 ############


def __ts_covers(two_symbol, permut_indexes, input, verbose=False):
    """Helper method to test if an input is being covered by a two symbol permuted implicant

    Args:
        two_symbol (string): the two_symbol implicant.
        permut_indexes (list): a list-of-lists of the implicant indexes that are permutables.
        input (string): the input string to be checked.

    Returns:
        x (bool): True if covered else False.
    """
    # No permutation, just plain implicant coverage?
    if not len(permut_indexes):
        if __pi_covers(two_symbol, input):
            return True
    # There are permutations to generate and check
    else:
        # NEW METHOD: Generates the expanded logic of the Two-Symbol Schema
        for gen_implicant in _expand_ts_logic(two_symbol, permut_indexes):
            if __pi_covers(gen_implicant, input):
                return True
        """
        # OLD METHOD
        for idxs in permut_indexes:
            # Extract the charactes that can be permuted
            chars = [implicant[idx] for idx in idxs]
            # Generate all possible permutations of these symbols
            permut_chars = itertools.permutations(chars, len(idxs))
            for permut_chars in permut_chars:
                # Generate a new implicant and substitute the charactes with the permuted ones
                tmp = list(implicant)
                for idx,char in zip(idxs,permut_chars):
                    tmp[idx] = char
                # The new permuted implicate is covered?
                if __pi_covers(tmp, input):
                    return True
        """
    return False


def computes_ts_coverage(k, outputs, two_symbols):
    """ Computes the input coverage by Two Symbol schematas.

    Args:
        k (int): the number of inputs.
        outpus (list): the list of transition outputs.
        two_symbols (list): The final list of Two Symbol permutable schematas. This is returned by `find_two_symbols`.

    Returns:
        coverage (dict): a dictionary of coverage where keys are inputs states and values are lists of the Two Symbols covering that input.
    """
    ts_coverage = {}
    for statenum in range(2**k):
        binstate = statenum_to_binstate(statenum, base=k)
        ts_coverage[binstate] = covering_twosymbols = []
        output = int(outputs[statenum])
        if output == 2:
            output = [0, 1]
        else:
            output = [int(outputs[statenum])]
        for t in output:
            for implicant, permut_indxs, same_symbols_indxs in two_symbols[t]:
                if __ts_covers(implicant, permut_indxs, binstate):
                    covering_twosymbols.append((implicant, permut_indxs, same_symbols_indxs))
    #
    return ts_coverage
